@@ -1,11 +1,25 @@
+from tkinter import simpledialog
+from tkinter.ttk import Notebook
 from ScrollText import *
 from tkinter import *
 from tkinter import messagebox
 import socket, threading,select
 import SettingsDialog
+from wgplib import *
 
 
 class TerminalWindow(ScrollText):
+    pages:Notebook=None
+    getstack:bool=False
+    setstackvisible=None
+    stack:Text=None
+    errorverb:str=""
+    arglist={}
+    verblist=[]
+    namelist={}
+    lastobj:str=""
+    lvVerbs:ttk.Treeview=None
+
     ColorTable = (
         "#000000",
         "#c00000",
@@ -304,12 +318,12 @@ class TerminalWindow(ScrollText):
         self.bottom.pack(side=BOTTOM, fill=X)
         self.sendbtn = Button(self.bottom, text="Snd", command=self.doSend)
         self.mytext = StringVar()
-        self.sendcmd = Entry(self.bottom, textvariable=self.mytext)
+        self.sendEntry = Entry(self.bottom, textvariable=self.mytext)
         self.sendbtn.pack(side=RIGHT)
-        self.sendcmd.pack(fill=X, expand=True)
-        self.sendcmd.bind("<Return>",self.doSendEvent)
-        self.sendcmd.bind("<Up>", self.history)
-        self.sendcmd.bind("<Down>", self.history)
+        self.sendEntry.pack(fill=X, expand=True)
+        self.sendEntry.bind("<Return>",self.doSendEvent)
+        self.sendEntry.bind("<Up>", self.history)
+        self.sendEntry.bind("<Down>", self.history)
         self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 
         self.iscsi = False
@@ -329,6 +343,9 @@ class TerminalWindow(ScrollText):
         self.capturemode=0
         self.capturefunc=None
         self.external_edit=""
+        self.verbcollect=''
+        self.onExamineLine=None
+
         self.historylst=[]
         self.historyidx=0
         ifile=SettingsDialog.getConfig()
@@ -346,6 +363,21 @@ class TerminalWindow(ScrollText):
             self.textbox.tag_config(tagname, cfg)
             self.taglist.append(tagname)
         self.currenttag = tagname
+    
+    def parseVerb(self,value):
+        """Return (obj,verb), or False if not valid."""
+        if value.startswith('@'):
+             value=parse(value)[1]
+        (obj,value)=parsesep(value,':')
+        (verb,value)=parse(value)
+        if verb.startswith('"'):
+             verb=getsepfield(verb,1,'"')
+        i=verb.find("*")
+        if (i>=0):
+            verb=verb[0:i]
+        if (verb!='') and (obj!=''):
+            return (obj,verb)
+        return False
 
     def addtext(self, msg):
         for c in msg:
@@ -494,6 +526,8 @@ class TerminalWindow(ScrollText):
                 self.external_edit=self.lastline
                 self.capturemode=1
                 self.capturestr=""
+            elif self.onExamineLine:
+                self.onExamineLine(self.lastline)
         elif self.capturemode==1:
             if self.capturestr!="":
                 self.capturestr+="\n"
@@ -503,6 +537,209 @@ class TerminalWindow(ScrollText):
                 (name,upload)=self.parseExternal()
                 self.capturefunc(self.capturestr,name)
 
+    def selectError(self,obj:str,verb:str,lno:int)->bool:
+        tabs=self.pages.tabs()
+        for i in range(2,len(tabs)):
+            w=self.pages.nametowidget(tabs[i])
+            if (type(w) is ScrollText):
+                re:Text=w.textbox
+                lines=re.get("1.0","end").splitlines()
+                if (len(lines)>0 and (lines[0]+" ").lower().find(obj+':'+verb+' ')>=0):
+                    self.pages.select(i)
+                    start=str(lno+1)+'.0'
+                    end=str(lno+1)+".end"
+                    re.see(start)
+                    re.focus_set()
+                    re.tag_add(SEL,start,end)
+                    #pagesChange(self);
+                    return True
+        return False
+    
+    def currentEditor(self)->Text:
+        w=self.currentPage()
+        if (type(w) is ScrollText):
+            return w.textbox
+        return None
+
+    def currentPage(self)->ScrollText:
+        w=self.nametowidget(self.pages.select())
+        if (type(w) is ScrollText):
+            return w
+        return None
+
+    def addtarget(self,dest:Text, line:str):
+        dest.insert("end",line+"\n")
+        dest.see("end")
+
+    def doCheckTest(self,line:str):
+        """Check for stack trace messages"""
+        if self.getstack: 
+            self.addtarget(self.stack,line)
+        # #540:test (this == #540), line 5:  Type mismatch (expected integer; got float)
+        # #151:+attacks, line 9:  Verb not found: #548:energy_cast()
+        # #151:+deploy deploy, line 28:  Range error
+        if line.startswith('#') and line.find(', line')>=0:
+            startline=line
+            (prog,line)=parsesep(line,',')
+            line=parse(line)[1] # Skip "line"
+            (lno,line)=parsesep(line,':')
+            (obj,verb)=self.parseVerb(prog)    #Should strip out trailing defs.
+            error=line
+            self.getstack=True
+            self.stack.delete("1.0","end")
+            self.addtarget(self.stack,'Stack')
+            self.addtarget(self.stack,error)
+            self.addtarget(self.stack,obj+':'+verb+', line '+lno)
+            self.setstackvisible(True)
+            self.errorverb=''
+            if not self.selectError(obj,verb,int(lno)):
+                self.errorverb=verb
+                self.errorobj=obj
+                self.lastlno=int(lno)
+        elif (line=='(End of traceback)'):
+            self.getstack=False
+            if self.errorverb!='':
+                self.findVerb(self.errorobj,self.errorverb,self.lastlno)
+            self.errorverb=''
+    
+    def doCheckCompile(self,line:str)->None:
+        """Check that verb has compiled"""
+#        var lno,x,n:Integer; s1,s2:String;  e:TEdit;
+#        adddebug(line);
+        if (line.startswith('Line ')):
+            (s1,line)=parsesep(line,':');
+            lno=atol(s1[4:])
+            self.stack.delete("1.0","end")
+            self.stack.insert("1.0","Compile Error\n"+line)
+            self.setstackvisible(True)
+            re=self.currentEditor()
+            ix=str(lno+1)+".0"
+            re.mark_set(INSERT,ix)
+            re.see(ix)
+            re.tag_add(SEL,ix,str(lno+1)+".end")
+            re.focus_set()
+#            adddebug('Error found: '+line);
+        elif line=='Verb not programmed.':
+            self.onExamineLine=self.doCheckTest
+        elif line=='Verb programmed.':
+#            adddebug('Compiled OK.');
+            self.onExamineLine=self.doCheckTest
+            if False:
+#            SetChanged(CurrentEditor,false);
+#            e:=CurrentTest;
+#            if assigned(e) and (trim(e.Text)<>'') then
+#            begin
+#            msgqueue.add(e.Text);
+#            testtab:=pages.ActivePage;
+#            onExamineLine:=DoChecktest;
+                self.getstack=False
+                self.pages.select(self)
+            else:
+                 self.showmessage(line)
+
+    def addTab(self,caption:str,text:str,tabtype:int):
+        t=ScrollText(self.pages,background="black",foreground="white",font=("Courier",12,"bold"),insertbackground="white")
+        self.pages.add(t,text=caption)
+        t.textbox.insert("1.0",text)
+        t.tabtype=tabtype
+
+    def doCheckVerb(self, line:str):
+        if (line=='***finished***'):
+            self.onExamineLine=self.doCheckTest
+            t=self.verbcollect.splitlines()
+            t.append(".")
+            prog=t[0]
+            if (prog=='That object does not define that verb.'):
+                messagebox.showwarning("MooCoderPy",'Verb not found.')
+                return
+            for i in range(len(t)):
+                if t[i].endswith('[normal]'): # Stupid ansi is stupid.
+                    s=t[i]
+                    t[i]=s[0:len(s)-len('[normal]')]
+            (obj,prog)=parsesep(prog,':')
+            args=prog;
+            if (args.startswith('"')):
+                args=args.replace('"','').strip()
+            (verb,prog)=parse(prog)
+            if (verb.startswith('"')):
+                verb=getsepfield(verb,1,'"')
+            i=verb.find('*')
+            if (i>=0):
+                verb=verb[0:i]
+            t[0]='@program '+obj+':'+verb
+            if self.selectError(obj,verb,self.lastlno):
+                re=self.currentEditor()
+                re.delete("1.0","end")
+                re.insert("1.0","\n".join(t))
+                re.see(str(self.lastlno)+":0")
+            else:
+                self.addTab(obj+':'+verb,"\n".join(t),1)
+                self.selectError(obj,verb,self.lastlno)
+            self.lastlno=0
+            self.checkName(obj)
+            self.arglist[obj+':'+verb]=args
+            self.verblist.append(obj+':'+verb)
+            self.updateVerbs()
+        else:
+            self.verbcollect+=line+"\n"
+    
+    def checkName(self,obj:str)->None:
+        """Check that we know the name of an object."""
+        if not(obj in self.namelist):
+            self.sendCmd(';'+obj+'.name')
+            self.onExamineLine=self.doCheckName
+        self.lastobj=obj
+    
+    def doCheckName(self,line:str):
+        """Examine stream for name of object."""
+        if line.lower().find('***finished***')>=0:
+            return # Ignore trailing stuff.
+        if line.startswith('=> 0'):
+            return # tailing result.
+        self.onExamineLine=self.doCheckTest
+        if not(line.startswith('=>')):
+            self.addln('Name not found.')
+            return
+        aname=getsepfield(line,1,'"')
+        self.namelist[self.lastobj]=aname
+        self.updateVerbs
+    
+    def findVerbHelp(self,obj:str,verb:str)->str:
+        return ""
+
+    def updateVerbs(self)->None:
+        """Update Verb list window"""
+#        var nd:TListItem; s:String; obj,verb:String;
+#            oldverb,oldobj:String; i:Integer;
+#        lvVerbs.Items.BeginUpdate;
+        oldverb=''
+        nd=self.lvVerbs.item(self.lvVerbs.focus())
+        if nd["text"]!="":
+            oldobj=nd["text"].lower()
+            oldverb=nd["values"][1].lower()
+        for i in self.lvVerbs.get_children(): #Clear list
+            self.lvVerbs.delete(i)
+        for s in self.verblist:
+            obj=getsepfield(s,0,':')
+            verb=getsepfield(s,1,':')
+            name=self.namelist[obj] if obj in self.namelist else obj
+            values=(name,verb,self.arglist[s],self.findVerbHelp(obj,verb))
+            self.lvVerbs.insert("","end",text=obj,values=values)
+            self.fitListContents(self.lvVerbs)
+        if oldverb!="":
+            for i in self.lvVerbs.get_children():
+                nd=self.lvVerbs.items(i)
+                if nd["text"].lower()==oldobj and nd["values"][1].lower()==oldverb:
+                    self.lvVerbs.selection_set(i)
+                    self.lvVerbs.see(i)
+                    break
+
+    def fitListContents(self,alist:ttk.Treeview):
+        cols=alist.cget("columns")
+        colnames=["#0"]+list(cols)
+        print("cols",cols,colnames)
+        
+
     def parseExternal(self):
         try:
             t=self.external_edit.split(" name: ")[1]
@@ -520,6 +757,12 @@ class TerminalWindow(ScrollText):
         self.sendCmd(text)
         self.after(500,self.sendtext,"Update complete\n")
         return True
+    
+    def docompile(self,page:ScrollText):
+        if not(type(page) is ScrollText):
+            return
+        self.sendCmd(page.textbox.get("1.0","end"))
+        self.onExamineLine=self.doCheckCompile
 
     def doSend(self):
         s = self.mytext.get()
@@ -544,6 +787,29 @@ class TerminalWindow(ScrollText):
         if (self.connectString!=""):
             self.after(1000,self.sendCmd,self.connectString)
         print("Connected")
+    def showmessage(self,msg):
+        messagebox.showinfo("MooCoderPy",msg)
+
+    def loadVerb(self,verbdef):
+        v=self.parseVerb(verbdef)
+        if not v:
+            self.showmessage("Invalid Verb Syntax")
+            return
+        (obj,verb)=v
+        self.findVerb(obj,verb,0)
+
+    def findVerb(self,obj:str,verb:str, lno:int):
+        self.lastlno=lno
+        if (obj=='#-1'):
+            return # Not a real verb.
+        #if not(self.selectError(obj,verb,lno)):
+        self.fetchVerb(obj,verb)
+    
+    def fetchVerb(self,obj,verb):
+        self.sendCmd('@list '+obj+':'+verb+' without numbers')
+        self.sendCmd(';player:tell("***finished***")')
+        self.verbcollect=''
+        self.onExamineLine=self.doCheckVerb;
 
     def sendCmd(self,cmd):
         buf=(cmd+"\n").encode("utf-8")
